(ns puppetlabs.rbac.storage
  (:import [java.sql Timestamp])
  (:require [clojure.java.jdbc :as jdbc]
            [clojure.string    :as s]
            [clojure.set :refer [rename-keys difference]]
            [puppetlabs.kitchensink.core :as ks]
            [clj-time.coerce :as coerce]
            [slingshot.slingshot :refer [throw+]]))

;; TODO move these into jdbc-util
(defn- convert-result-arrays
  "Converts Java and JDBC arrays in a result set using the provided function
   (eg. vec, set). Values which aren't arrays are unchanged."
  ([result-set]
     (convert-result-arrays vec result-set))
  ([f result-set]
     (let [convert #(cond
                     (ks/array? %) (f %)
                     (isa? (class %) java.sql.Array) (f (.getArray %))
                     :else %)]
       (map #(ks/mapvals convert %) result-set))))

(defn- query
  "An implementation of query that returns a fully evaluated result (no
   JDBCArray objects, etc)"
  [db sql-and-params]
  (let [convert (fn [rs]
                  (doall
                   (convert-result-arrays vec
                                          (jdbc/result-set-seq rs))))]
    (jdbc/db-query-with-resultset db sql-and-params convert)))

(defn- insert! [db tbl data]
  ;; returns new record
  (first (into [] (jdbc/insert! db tbl data))))

(defn- update! [db tbl set-map where-clause]
  ;; returns number of rows updated
  (first (jdbc/update! db tbl set-map where-clause)))

(defn- in-query
  "Given a column name (either a string or keyword) and a list of
  values, returns a string of the form:

  <column> IN ( ?, ?, ... )

  suitable for appending to some existing SQL query.

  If values is empty, returns an empty string (which can still be
  appended to SQL as a no-op)."
  [col values]
  (if (empty? values)
    ""
    (format "%s IN ( %s )"
            (name col)
            (s/join ", " (map (constantly "?") values)))))

(defn compare-related-ids
  "This function determines the related ids that need to be added and
  removed for a given database object (e.g. a subject).

  id is the ID of the object whose related IDs we are concerned about.
  desired-ids is the list of up-to-date IDs we want reflected in the database.
  table is a keyword or string referring to the table housing the relationship.
  column-name is a string or keyword that refers to the column that would house the ID in question.

  This function should only be used with tables with two columns of ids.

  Examples:
  ; looks up subject where id = 1, finds its associated roles in the database
  ; and compares them to the desired list of roles
  (compare-related-ids db-spec 1 [1 2 3] :subject_roles :sid)
  ; look up subject where id = 1, finds its associated groups in the database
  ; and compares them to the desired list of groups
  (compare-related-ids db-spec 1 [4 5 6] :groupings :ruid)"
  [db id desired-ids table column-name]
  (let [table       (name table)
        column-name (name column-name)
        rows        (query db [(format "SELECT * FROM %s WHERE %s = ?" table column-name) id])
        db-ids      (set (map #(->> (keyword column-name)
                                    (dissoc %)
                                    (vals)
                                    (first))
                                rows))
        desired-ids (set desired-ids)]
    (let [to-add    (difference desired-ids db-ids)
          to-remove (difference db-ids desired-ids)]
      [to-add to-remove])))

;;;;;;

(defn dissoc-one-from-many! [db tbl one-col one many-col many]
  (when (seq many)
    (doseq [thing many]
      (jdbc/delete! db tbl [(format "%s = ? AND %s = ?" (name one-col) (name many-col)) one thing]))))

(defn assoc-one-to-many! [db tbl one-col one many-col many]
  (when (seq many)
    (apply jdbc/insert! db tbl [one-col many-col] (map (fn [thing] [one thing]) many))))

;;;;;;
(def subjects-select "SQL for selecting all subjects."
  "SELECT s.id AS sid,
          s.login,
          s.is_group,
          s.is_remote,
          s.is_superuser,
          s.display_name,
          s.email,
          s.is_revoked,
          s.last_login,
          g.gid AS gid,
          r.id AS rid,
          ir.id AS irid
   FROM subjects s
   LEFT OUTER JOIN groupings g ON (s.id = g.ruid)
   LEFT OUTER JOIN subject_roles sr ON (s.id = sr.sid)
   LEFT OUTER JOIN roles r ON (r.id = sr.rid)
   LEFT OUTER JOIN subject_roles gsr ON (g.gid = gsr.sid)
   LEFT OUTER JOIN roles ir ON (ir.id = gsr.rid)")

(def local-users-select "Tacks a WHERE clause onto subjects-select to select only local users."
  (s/join "\n" [subjects-select
                "WHERE s.is_group = false"
                "AND s.is_remote = false"]))

(def remote-users-select "Tacks a WHERE clause onto subjects-select to select only remote users."
  (s/join "\n" [subjects-select
                "WHERE s.is_group = false"
                "AND s.is_remote = true"]))

(def groups-select
  "Select all groups"

  "SELECT subjects.id,
          subjects.login,
          subjects.is_group,
          subjects.is_remote,
          subjects.is_superuser,
          subjects.display_name,
          agg_roles.role_ids,
          agg_groups.user_ids
  FROM subjects
  LEFT JOIN (
      SELECT gid, array_agg(ruid) AS user_ids
      FROM groupings
      GROUP BY gid
  ) agg_groups ON subjects.id = agg_groups.gid
  LEFT JOIN (
      SELECT sid, array_agg(rid) AS role_ids
      FROM subject_roles
      GROUP BY sid
  ) agg_roles ON subjects.id = agg_roles.sid
  WHERE subjects.is_group = true")

(def users-select
  (s/join "\n" [subjects-select
                "WHERE s.is_group = false"]))

(def roles-select "SQL for selecting all roles"
  "SELECT r.description,
          r.display_name AS role_name,
          r.id AS rid,
          g.id AS gid,
          u.id AS uid,
          p.id AS pid,
          p.object_type,
          p.action,
          p.instance
   FROM roles r
   LEFT OUTER JOIN subject_roles sr ON (r.id = sr.rid)
   LEFT OUTER JOIN subjects g ON (g.id = sr.sid AND g.is_group = true)
   LEFT OUTER JOIN subjects u ON (u.id = sr.sid AND u.is_group = false)
   LEFT OUTER JOIN roles_permissions rp ON (r.id = rp.rid)
   LEFT OUTER JOIN permissions p ON (p.id = rp.pid)")

(def permissions-select "SQL for selecting all permissions"
"SELECT  p.object_type, p.action, p.instance
 FROM subjects s
      LEFT JOIN groupings g ON (s.id = g.ruid)
      INNER JOIN subject_roles sr ON (s.id = sr.sid OR g.gid = sr.sid)
      INNER JOIN roles r ON (r.id = sr.rid)
      INNER JOIN roles_permissions rp ON (rp.rid = r.id)
      INNER JOIN permissions p ON (rp.pid = p.id)")

(defn create-permissions-query
  "Creates a WHERE clause with the subject id and the permissions query triple
  (:object_type, and optionally :action and/or :instance). If :instance is
  part of the query, looks for permissions matching the instance or *. Returns
  a vector with the first argument being the query string, and the subsequent
  arguments being the variables that get passed into this string."
  [perm-query sid]
  (cond-> [(str permissions-select " WHERE s.id = ?") sid]
    (:object_type perm-query) (-> (update-in [0] str " AND p.object_type = ?")
                                  (conj (:object_type perm-query)))
    (:action perm-query) (-> (update-in [0] str " AND p.action = ?")
                             (conj (:action perm-query)))
    (:instance perm-query) (-> (update-in [0] str " AND (p.instance = '*' OR p.instance = ?)")
                               (conj (:instance perm-query)))))

(defn get-permissions
  "Takes a db spec, a subject id, and a vector of permissions queries, and
  returns a vector of vectors of maps.

  `perm-query` is a vector of maps. Each map is a separate permission query
  that must contain :object_type and may contain :action and :instance.

  If the `perm-query` vector is empty, return all the permissions the subject has.
  Otherwise, map over each item in the vector, and return a vector of vectors
  of maps, where each inner vector is a list of permissions that match the
  query."
  ;; TODO handle a malformed query
  ;; TODO error if subject doesn't exist
  [db sid perm-query]
  (if (empty? perm-query)
    [(query db (create-permissions-query perm-query sid))]
    (mapv #(vec (query db (create-permissions-query % sid))) perm-query)))

;; The following extract-* and embed-* helpers are for massaging the
;; output of the *-select queries above into nested maps of related
;; things (subjects, roles, permissions).
(defn extract-subject [row]
  (-> (select-keys row [:sid :login :is_group :is_remote :is_superuser :display_name :email :is_revoked :last_login])
      (rename-keys {:sid :id})))

(defn extract-group [row]
  (-> (extract-subject row)
      (dissoc :email :is_revoked)))

(defn extract-role [row]
  (-> (select-keys row [:rid :role_name :description])
      (rename-keys {:role_name :display_name
                    :rid :id})))

(defn extract-perm [row]
  (-> (select-keys row [:rid :pid :object_type :action :instance])
      (rename-keys {:pid :id})))

(defn extract-rid [row]
  (-> (select-keys row [:sid :rid])
      (rename-keys {:rid :id})))

(defn extract-irid [row]
  (-> (select-keys row [:sid :irid])
      (rename-keys {:irid :id})))

(defn extract-gid [row]
  (-> (select-keys row [:sid :gid])
      (rename-keys {:gid :id})))

(defn extract-gid-for-role [row]
  (-> (select-keys row [:rid :gid])
      (rename-keys {:gid :id})))

(defn extract-uid [row]
  (-> (select-keys row [:rid :uid])
      (rename-keys {:uid :id})))

(defn extract-action [rows]
  (set
    (map (fn [row]
           (-> (select-keys row [:object_type :a_name :a_display_name :a_description :has_instances])
               (rename-keys {:a_name         :name
                             :a_display_name :display_name
                             :a_description  :description}))) rows)))

(defn extract-type [rows]
  (set
    (map (fn [row]
           (select-keys row [:object_type :display_name :description])) rows)))

(defn extract-with [fn rows]
  (->> rows
       (map fn)
       set ; potentially dedupe
       (filter #(not (nil? (:id %))))
       (into []))) ; back into persistent vector

(defn embed-perms-for-role [perms role]
  (let [relevant-perms (->> perms
                            (filter #(= (:id role) (:rid %)))
                            (mapv #(dissoc % :rid :id)))]
    (assoc role :permissions relevant-perms)))

(defn embed-roles-for-subject [roles subject]
  (let [relevant-roles (->> roles
                            (filter #(= (:id subject) (:sid %)))
                            (map :id))]
    (assoc subject :role_ids relevant-roles)))

(defn embed-inherited-roles-for-subject [roles subject]
  (let [relevant-roles (->> roles
                            (filter #(= (:id subject) (:sid %)))
                            (map :id))]
    (assoc subject :inherited_role_ids relevant-roles)))

(defn embed-ids-for-resource [result-key ids filter-key resource]
  (let [relevant-ids (->> ids
                          (filter #(= (:id resource) (filter-key %)))
                          (mapv :id))]
    (assoc resource result-key relevant-ids)))

(defn embed-groups-for-remote-user [groups user]
  (let [relevant-groups (->> groups
                             (filter #(= (:id user) (:sid %)))
                             (map :id))]
    (assoc user :group_ids relevant-groups)))

(defn embed-actions-for-types [actions object_type]
  (let [relevant-actions (mapv #(dissoc % :object_type) (filter #(= (:object_type object_type) (:object_type %)) actions))]
    (assoc object_type :actions relevant-actions)))

(defn timestamp->str
  "Convert `timestamp` to a string in ISO standard format"
  [timestamp]
  (-> timestamp
      coerce/to-date-time
      coerce/to-string))

(defn stringify-last-login
  "Convert the `subject`'s :last_login from
   a Timestamp object to the stringified version"
  [{:keys [last_login] :as subject}]
  (if last_login
    (assoc subject :last_login (timestamp->str last_login))
    subject))

(defn compose-local-users
  "Given the output rows generated by a local user select statement, build
  the complex objects (with related IDs) required by the front-end."
  [rows]
  (let [users (extract-with extract-subject rows)
        roles (extract-with extract-rid rows)]
    (map (comp stringify-last-login
               (partial embed-roles-for-subject roles)) users)))

(defn compose-remote-users
  "Given the output rows generated by a remote user select statement, build
  the complex objects (with related IDs) required by the front-end."
  [rows]
  (let [users           (extract-with extract-subject rows)
        roles           (extract-with extract-rid rows)
        groups          (extract-with extract-gid rows)
        inherited-roles (extract-with extract-irid rows)]
    (map
     (comp stringify-last-login
           (partial embed-roles-for-subject roles)
           (partial embed-groups-for-remote-user groups)
           (partial embed-inherited-roles-for-subject inherited-roles)) users)))

(defn compose-groups
  [rows]
    (let [groups (extract-with extract-group rows)
          roles  (extract-with extract-rid rows)]
      (map (partial embed-roles-for-subject roles) groups)))

(defn compose-roles
  "Given the output rows generated by get-role(s), build the nested
   objects required by the front-end."
  [rows]
    (let [roles    (extract-with extract-role rows)
          groups   (extract-with extract-gid-for-role rows)
          users    (extract-with extract-uid rows)
          perms    (extract-with extract-perm rows)]
      (map (comp (partial embed-perms-for-role perms)
                 (partial embed-ids-for-resource :group_ids groups :rid)
                 (partial embed-ids-for-resource :user_ids users :rid))
           roles)))

(defn compose-types [rows]
  (let [types   (extract-type rows)
        actions (extract-action rows)]
    (map (partial embed-actions-for-types actions) types)))

(defn- select-subjects
  "Fetches all subjects that match the provided column and value. Does
  no composition of related things (roles nor groups)."
  [db column value]
  (let [column (name column)]
    (query db [(str subjects-select (format " WHERE s.%s = ?" column)) value])))

(defn get-subject [db column value]
  (let [subjects (select-subjects db column value)
        subject  (first subjects)]
    (when (seq subjects)
      (let [compose-fn (cond
                        (:is_group subject) compose-groups
                        (:is_remote subject) compose-remote-users
                        :else compose-local-users)]
        (first (compose-fn subjects))))))

(defn get-users
  [db & [ids]]
  (let [ids                 (if (nil? ids) [] ids)
        in-query            (in-query :s.id ids)
        local-users-select  (if (empty? in-query) local-users-select (format "%s AND %s" local-users-select in-query))
        remote-users-select (if (empty? in-query) remote-users-select (format "%s AND %s" remote-users-select in-query))
        local-users         (-> (query db (vec (flatten [local-users-select ids])))
                                (compose-local-users))
        remote-users        (-> (query db (vec (flatten [remote-users-select ids])))
                                (compose-remote-users))]
    (concat local-users remote-users)))

(defn get-user [db column value]
  (when-let [subject (get-subject db column value)]
    (when-not (:is_group subject)
      subject)))

(defn get-groups [db & [ids]]
  (let [ids           (if (nil? ids) [] ids)
        in-query      (in-query :subjects.id ids)
        groups-select (if (empty? in-query) groups-select (format "%s AND %s" groups-select in-query))]
  (query db (vec (flatten [groups-select ids])))))

(defn get-group [db column value]
  (let [column (name column)
        [group] (query db [(str groups-select (format " AND subjects.%s = ?" column)) value])]
    group))

(defn get-roles
  "Fetches all roles and related objects (subjects, permissions) from
   the database using `roles-select`."
  [db & [ids]]
  (let [ids          (if (nil? ids) [] ids)
        in-query     (in-query :r.id ids)
        roles-select (if (empty? in-query) roles-select (format "%s WHERE %s" roles-select in-query))
        rows         (query db (vec (flatten [roles-select ids])))]
    (compose-roles rows)))

(defn get-role
  "Fetches a single role and its related objects (subjects,
   permissions) from the database using `roles-select` based on the
   provided column & value pair."
  [db column value]
  (let [column (name column)
        rows   (query db [(str roles-select (format " WHERE r.%s = ?" column)) value])]
    (when (seq rows)
      (first (compose-roles rows)))))

(defn update-associated-ids!
  "Updates, as necessary, the object represented by one-id to be
  associated with only the IDs specified by many-ids. one-col and
  many-col specify the columns to use to look up one-id and many-ids,
  respectively, and table is the table that holds the relationship
  between one-col and many-col."
  [db table one-col one-id many-col many-ids]
  (jdbc/with-db-transaction [db-conn db]
    (let [[ids-to-assoc
           ids-to-dissoc] (compare-related-ids db-conn one-id many-ids table one-col)]
      (assoc-one-to-many! db-conn table one-col one-id many-col ids-to-assoc)
      (dissoc-one-from-many! db-conn table one-col one-id many-col ids-to-dissoc))))

(defn create-subject! [db data]
  (let [set-map (dissoc data :role_ids)
        rids    (:role_ids data)]
    (jdbc/with-db-transaction [db-conn db]
      (let [{sid :id} (insert! db :subjects set-map)]
        (assoc-one-to-many! db-conn :subject_roles :sid sid :rid rids)))))

(defn create-subject-with-flags [& flags]
  (let [defaults {:is_group false :is_superuser false :is_remote false}
        flag-map (merge defaults (zipmap flags (repeat true)))]
    (fn [db data]
      (create-subject! db (merge data flag-map)))))

(def create-local-user!  (create-subject-with-flags))
(def create-remote-user! (create-subject-with-flags :is_remote))
(def create-group!       (create-subject-with-flags :is_group :is_remote))

(def cached-admin-user (atom nil))

(defn admin-user [db]
  (if-let [user @cached-admin-user]
    user
    (do
      (reset! cached-admin-user (get-subject db "login" "admin"))
      @cached-admin-user)))

(defn check-admin-user-update
  "Checks to ensure if the user is updating the admin user, then none of
   the immutable fields are included in the update. If so an exception
   is thrown, otherwise just returns nil"
  [admin to-be-updated-id updated-data]
  (when (= to-be-updated-id (:id admin))
    (let [throw-map {:kind ::admin-user-immutable}]
      (cond
       (not= (:login updated-data)
             (:login admin))
       (throw+ (assoc throw-map :msg "Admin user login cannot be changed"))

       (not= (:is_revoked updated-data)
             (:is_revoked admin))
       (throw+ (assoc throw-map :msg "Admin user cannot be revoked"))

       (not= (set (:role_ids updated-data))
             (set (:role_ids admin)))
       (throw+ (assoc throw-map :msg "Cannot change the admin user's roles"))))))

(defn update-local-user! [db id data]
  (let [to-update (select-keys data [:login
                                     :display_name
                                     :email
                                     :is_revoked])
        rids    (:role_ids data)]

    (check-admin-user-update (admin-user db) id data)

    (jdbc/with-db-transaction [db-conn db]
      (update! db-conn :subjects to-update ["id = ?" id])
      (update-associated-ids! db-conn :subject_roles :sid id :rid rids))))

(defn update-remote!
  "For use in the API. Updates role_ids and is_revoked, for either groups or
  remote users."
  [db id {role_ids :role_ids is_revoked :is_revoked}]
  (jdbc/with-db-transaction [db-conn db]
    (update! db-conn :subjects {:is_revoked is_revoked} ["id = ?" id])
    (update-associated-ids! db-conn :subject_roles :sid id :rid role_ids)))

(defn update-user! [db id data]
  (if (:is_remote data)
    (update-remote! db id data)
    (update-local-user! db id data)))

(def perm-select
  "SELECT id FROM permissions WHERE object_type = ? AND action = ? AND instance = ?")

(defn valid-perm?
  "Given a perm, ensure that the object_type and action exist in the
  database already (by looking up one of the default permissions) and
  that instance is not nil."
  [db perm]
  (and
   (not (empty? (query db [perm-select (:object_type perm) (:action perm) "*"])))
   (not (empty? (:instance perm)))))

(defn perm->id!
  "Given a permission map, either extract an id, fetch an id for the
  perm based on its keys, or create the perm and return the new ID."
  [db perm]
  (if-let [pid (:id perm)]
    pid
    (if-let [pid (-> (query db [perm-select (:object_type perm) (:action perm) (:instance perm)])
                     first
                     :id)]
      pid
      (if (valid-perm? db perm)
        (:id (insert! db :permissions perm))
        (throw (IllegalArgumentException. (format "Invalid permission: %s" perm)))))))

(defn create-role! [db data]
  (jdbc/with-db-transaction [db-conn db]
    (let [to-insert            (select-keys data [:display_name :description])
          {perms :permissions
           gids  :group_ids
           uids  :user_ids} data
          pids                 (mapv (partial perm->id! db-conn) perms)]
      (let [{rid :id} (insert! db :roles to-insert)]
        (assoc-one-to-many! db-conn :roles_permissions :rid rid :pid pids)
        (assoc-one-to-many! db-conn :subject_roles :rid rid :sid gids)
        (assoc-one-to-many! db-conn :subject_roles :rid rid :sid uids)))))

(def immutable-roles* (atom nil))

(def immutable-role-names #{"Administrators" "Operators" "Viewers"})

(defn perm-id-only [db role]
  (update-in role [:permissions]
             (fn [perms]
               (set (map #(perm->id! db %) perms)))))

(defn create-immutable-role-map [db]
  (reduce (fn [acc {:keys [id] :as role}]
            (assoc acc id role))
          {} (for [role (get-roles db)
                   :when (contains? immutable-role-names (:display_name role))]
               (perm-id-only db role))))

(defn immutable-role
  "Check `role-id` against known immutable roles. If found, return
   the immutable details of the role"
  [db role-id]
  (if-let [roles @immutable-roles*]
    (get roles role-id)
    (do
      (reset! immutable-roles* (create-immutable-role-map db))
      (get @immutable-roles* role-id))))

(defn check-immutable-role
  "Throws an exception if `id` is an immutable (default) role"
  [db id {new-name :display_name
          new-description :description} perm-ids]
  (when-let [{:keys [display_name description permissions]} (immutable-role db id)]
    (cond
     (not= new-name display_name)
     (throw+ {:kind ::default-roles-immutable
              :msg (format "Cannot modify display name for default role %s" display_name)})

     (not= new-description description)
     (throw+ {:kind ::default-roles-immutable
              :msg (format "Cannot modify description for default role %s" display_name)})

     (not= (set perm-ids) permissions)
     (throw+ {:kind ::default-roles-immutable
              :msg (format "Cannot modify permissions for default role %s" display_name)})

     :else
     nil)))

(defn administrator-role?
  [db id]
  (= "Administrators" (:display_name (immutable-role db id))))

(defn check-admin-in-admin-role
  [db id user-ids]
  (let [admin-id (-> db admin-user :id)]
    (when (and (administrator-role? db id)
               (not-any? #(= admin-id %) user-ids))
      (throw+ {:kind ::admin-user-not-in-admin-role
              :msg "Cannot remove the admin user from the 'Administrators' role"}))))

(defn update-role! [db id data]
  (jdbc/with-db-transaction [db-conn db]
    (let [uids    (:user_ids data)
          gids    (:group_ids data)
          perms   (:permissions data)
          pids    (map (partial perm->id! db-conn) perms)
          set-map (dissoc data :id :permissions :user_ids :group_ids)]

      (check-immutable-role db id set-map pids)
      (check-admin-in-admin-role db id uids)

      (update! db-conn :roles set-map ["id = ?" id])
      (update-associated-ids! db-conn :subject_roles :rid id :sid (concat uids gids))
      (update-associated-ids! db-conn :roles_permissions :rid id :pid pids))))

(defn delete-role! [db id]
  (jdbc/delete! db :roles ["id = ?" id]))

(defn get-types [db]
  (let [rows (query db ["SELECT ot.*,
                                a.name AS a_name,
                                a.display_name AS a_display_name,
                                a.description AS a_description,
                                a.has_instances as has_instances
                        FROM object_types ot
                        LEFT OUTER JOIN actions a ON (ot.object_type = a.object_type)"])]
    (compose-types rows)))

(defn get-directory-settings [db]
  (let [result (-> (query db ["select * from directory_settings"])
                   first
                   (dissoc :id))]
    (if (every? nil? (vals result))
      {}
      result)))

(defn update-directory-settings! [db data]
  (if (seq data)
    (update! db :directory_settings data ["id = 1"])
    (jdbc/with-db-transaction [db-conn db]
      (jdbc/delete! db-conn :directory_settings ["id = 1"])
      (jdbc/insert! db-conn :directory_settings {:id 1}))))

(defn update-user-password! [db login new-password clear-token-uuid?]
  (let [col-map (when clear-token-uuid?
                  {:reset_password_uuid nil})]
    (jdbc/update! db :subjects (assoc col-map
                                 :password new-password
                                 :failed_login_attempts 0
                                 :is_revoked false) ["login=?" login])))

(defn password-matches? [db login old-password]
  (boolean (first (query db ["SELECT * FROM subjects WHERE login=? AND password=?" login old-password]))))

(defn mark-user-for-reset [db sid]
  (let [uuid (java.util.UUID/randomUUID)]
    (update! db :subjects {:reset_password_uuid uuid} ["id=?" sid])
    (str uuid)))

(defn current-timestamp []
  (java.sql.Timestamp. (System/currentTimeMillis)))

(defn update-last-login [db login]
  (jdbc/update! db :subjects {:last_login (current-timestamp)
                              :failed_login_attempts 0} ["login=?" login]))

(defn update-failed-login [db login max-failed-attempts]
  (jdbc/db-do-prepared db "UPDATE subjects
                            SET failed_login_attempts = failed_login_attempts + 1,
                                is_revoked = failed_login_attempts + 1 >= ?
                            WHERE login = ?" [max-failed-attempts login]))
